"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This script contains all functions to export the results to excel.
Script version date: 23.11.2024
"""  

""" Import packages """
import os
import xlwt
import numpy as np
import pandas as pd

""" Functions """
def save_excel(config, ROOT_DIR, NAME_EXCEL_FILE, holdup_packing, average_contacts, contact_probabilities, q0_initial, q0_rlp, q0_final, q3_initial, q3_rlp, q3_final, d_classes_final, n_particle_final_rlp, n_particle_nowall, n_particle_final_gs):
    """ Function to export results data to excel file """
    # Convert the arrays into dataframes
    holdup_packing_df = pd.DataFrame(100*holdup_packing)
    average_contacts_df = pd.DataFrame(average_contacts)
    contact_probabilities_df = pd.DataFrame(contact_probabilities)
    q0_initial_df = pd.DataFrame(100*q0_initial)
    q0_rlp_df = pd.DataFrame(100*q0_rlp)
    q0_final_df = pd.DataFrame(100*q0_final)
    q3_initial_df = pd.DataFrame(100*q3_initial)
    q3_rlp_df = pd.DataFrame(100*q3_rlp)
    q3_final_df = pd.DataFrame(100*q3_final)
    # Round class diameters 
    np.round(d_classes_final, decimals=0, out=d_classes_final)
    d_classes_final_df = pd.DataFrame(d_classes_final)
    # Define headers
    headers_simulation = ['Seed', 'Final number of particles']
    headers_holdup_packing = ['Grow step', 'Hold-up']
    headers_d_classes = ['Class', 'Class diameter']
    header_average_contacts = ['Grow step']
    headers_dsd = ['q0_initial', 'q0_rlp', 'q0_final', 'q3_initial', 'q3_rlp', 'q3_final']
    for i in range(config.n_classes):
        header_average_contacts.append(f'Class {i+1}')
    header_contact_probabilities = []
    for i in range(config.n_classes):
        header_contact_probabilities.append(f'Class {i+1}')
    # Create a new Excel workbook
    workbook = xlwt.Workbook()
    # Create a new Excel sheet for the holdup data
    sheet_main = workbook.add_sheet('Main')
    # Write simulation settings headers to Excel
    for row_index, header in enumerate(headers_simulation):
        sheet_main.write(row_index, 0, header)
    # Write simulation settings data to Excel
    sheet_main.write(0, 1, config.seed)
    sheet_main.write(1, 1, n_particle_final_rlp)
    sheet_main.write(1, 2, n_particle_nowall)
    sheet_main.write(1, 3, n_particle_final_gs)
    # Write holdup headers to Excel
    for row_index, header in enumerate(headers_holdup_packing):
        sheet_main.write(row_index + 3, 0, header)
    # Write holdup data to Excel
    for col_index, value in enumerate(holdup_packing_df.values):
        sheet_main.write(3, col_index + 1, col_index)
        sheet_main.write(4, col_index + 1, value[0])
    # Write classes headers to Excel
    for row_index, header in enumerate(headers_d_classes):
        sheet_main.write(row_index + 6, 0, header)
    # Write classes data to Excel
    for col_index in range(len(d_classes_final)):
        sheet_main.write(6, col_index + 1, col_index + 1)
    for col_index, value in enumerate(d_classes_final_df.values):
        sheet_main.write(7, col_index + 1, value[0])
    # Write DSD headers to Excel
    for row_index, header in enumerate(headers_dsd):
        sheet_main.write(row_index + 9, 0, header)
    # Write DSD data to Excel
    for col_index, value in enumerate(q0_initial_df.values):
        sheet_main.write(9, col_index + 1, value[0])
    for col_index, value in enumerate(q0_rlp_df.values):
        sheet_main.write(10, col_index + 1, value[0])
    for col_index, value in enumerate(q0_final_df.values):
        sheet_main.write(11, col_index + 1, value[0])
    for col_index, value in enumerate(q3_initial_df.values):
        sheet_main.write(12, col_index + 1, value[0])
    for col_index, value in enumerate(q3_rlp_df.values):
        sheet_main.write(13, col_index + 1, value[0])
    for col_index, value in enumerate(q3_final_df.values):
        sheet_main.write(14, col_index + 1, value[0])
    # Create a new Excel sheet for the contacts data
    sheet_contacts = workbook.add_sheet('Contacts')
    # Write contacts headers to Excel
    for col_index, header in enumerate(header_average_contacts):
        sheet_contacts.write(0, col_index, header)
    # Write contacts data to Excel
    for row_index, values in enumerate(average_contacts_df.values):
        sheet_contacts.write(row_index + 1, 0, row_index)
        for col_index, value in enumerate(values):
            sheet_contacts.write(row_index + 1, col_index + 1, value)
    # Iterate through the nested dataframe of contact probabilities
    for sheet_index, sheet_data in enumerate(contact_probabilities_df.values):
        # Create a new Excel sheet for each inner list
        sheet_probabilities = workbook.add_sheet(f'Probabilities Step {sheet_index}')  
        # Write probabilities headers to Excel
        for col_index, header in enumerate(header_contact_probabilities):
            sheet_probabilities.write(0, col_index + 1, header)
            sheet_probabilities.write(col_index + 1, 0, header)
        # Iterate through each sublist 
        for row_index, sublist in enumerate(sheet_data):
            # Write contact probabilities to Excel
            for col_index, value in enumerate(sublist):
                sheet_probabilities.write(row_index + 1, col_index + 1, value)
    # Save the workbook to a file
    file_name = os.path.join(ROOT_DIR, NAME_EXCEL_FILE)
    workbook.save(file_name)