"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This script contains all functions related to the droplet grow algorithm.
Script version date: 23.11.2024
"""  

""" Import packages """

import numpy as np
from tess import Container
from scipy.spatial import Delaunay
from scipy.optimize import minimize

""" Functions """
def voro_tess(M_particle_tess, tess_start, tess_end, diameters_0): 
    """ Function to perform Tesselation """
    cntr = Container(M_particle_tess, limits=(tess_start, tess_end), radii=(diameters_0/2), periodic=False)
    # Cell volumes
    volumes_cells = [round(v.volume(), 3) for v in cntr]
    # Cell neighbors
    cell_neighbors = [v.neighbors() for v in cntr]
    # Nested tuple structure with x,y,z coordinates of the cell face normals in form:
    # [[(x,y,z), (x,y,z), ...], [(x,y,z), (x,y,z), ...], ...]
    cell_face_normals = [v.normals() for v in cntr]
    # Nested tuple structure with x,y,z coordinates of the cell verticies in form:
    # [[(x,y,z), (x,y,z), ...], [(x,y,z), (x,y,z), ...], ...]
    cell_vertices = [v.vertices() for v in cntr]
    # A list of the indices of the vertices of each face.
    vertices_indices = [v.face_vertices() for v in cntr]

    return volumes_cells, cell_neighbors, cell_face_normals, cell_vertices, vertices_indices

def calc_grow(config, particle_i, volumes_particles, diameters, cell_face_fixed, distances_faces_max, cell_face_normals, vertices_indices, cell_vertices, particle_status, cell_status):
    """ Function to simulate particle grow """
    # Particle volume
    volume_particle_i = volumes_particles[particle_i]
    # Particle radius
    radius_particle_i = diameters[particle_i] / 2
    # List of fixed faces of the considered particle
    cell_face_fixed_i = cell_face_fixed[particle_i]
    # Verticies of the cell corresponding to the particle
    cell_vertices_i = cell_vertices[particle_i]
    # List of maximal distances between particle center and cell faces for the considered particle
    distances_faces_max_i = distances_faces_max[particle_i]
    # Normal vectors of cell faces corresponding to the particle
    cell_face_normals_i = cell_face_normals[particle_i]
    # Indices of the vertices of each cell face corresponding to the particle.
    vertices_indices_i = vertices_indices[particle_i]
    # Cell status
    cell_status_i = cell_status[particle_i]
    # Initial guess for solver
    initial_guess = radius_particle_i 
    # Bounds
    bnds = [(0, np.max(distances_faces_max_i))]
    # Solver method
    solver_method = 'SLSQP'
    if cell_status_i == False:
        # Perform the optimization
        result = minimize(solver_distances, initial_guess, args=(config, cell_face_fixed_i, distances_faces_max_i, volume_particle_i, cell_face_normals_i, vertices_indices_i, cell_vertices_i, radius_particle_i), method=solver_method, bounds=bnds, tol=1e-3)
        # Obtain the value of the minimized function
        minimized_value = result.fun
        # If value to high, mark particle as invalid
        if minimized_value > 1.0:
            particle_status[particle_i] = False
        # Extract the distance
        distance_face_var = result.x[0]
        # Iterate through each face of the cell to update the list of fixed cell faces
        for face in range(len(cell_face_normals_i)):
            if cell_face_fixed[particle_i][face] == True:
                continue
            else:
                # If the distance is less than or equal to the radius of the sphere, the face intersects with the sphere. 
                # Thus it is set to fixed
                if (np.abs(distance_face_var) - distances_faces_max[particle_i][face]) > 1e-9:
                    cell_face_fixed[particle_i][face] = True  
    else:
        # Iterate through each face of the cell to update the list of fixed cell faces
        for face in range(len(cell_face_normals_i)):
            if cell_face_fixed[particle_i][face] == True:
                continue
            else:
                cell_face_fixed[particle_i][face] = True

    return cell_face_fixed, particle_status

def solver_distances(fit_pars, config, cell_face_fixed_i, distances_faces_max_i, volume_particle_i, cell_face_normals_i, vertices_indices_i, cell_vertices_i, radius_particle_i):
    """ Solver to calculate the distance between the particle center and the variable cell faces """
    # Variable distance from particle center to the cell face
    distance_face_var = fit_pars
    # Minimization value
    total_error = 0
    # Calculate new verticies
    cell_vertices_i_new = calc_verticies(config, cell_face_normals_i, vertices_indices_i, distances_faces_max_i, distance_face_var, cell_face_fixed_i, cell_vertices_i, radius_particle_i)
    # Check arrayfor NaN values
    if np.isnan(cell_vertices_i_new).any():
        total_error = 10000
    else:
        # Calculate polyhedron volume 
        volume_polyhedron = calc_volume_polyhedron(cell_vertices_i_new)
        # Minimization function
        total_error = 100*abs(volume_particle_i - volume_polyhedron) / volume_particle_i

    return total_error

def calc_verticies(config, cell_face_normals_i, vertices_indices_i, distances_faces_max_i, distance_face_var, cell_face_fixed_i, cell_vertices_i, radius_particle_i):
    """ Function to calculate new verticies coordinates """
    # Predefine list with aim distances between particle center and cell faces
    distances_faces = []
    # Chech each face of polyhedron
    for face_index in range(len(cell_face_normals_i)):
        # If face is from beginning fixed set distance to particle radius
        if cell_face_fixed_i[face_index] == True:
            distance_face = distances_faces_max_i[face_index]
        # If the distance of variable face is bigger than the distance between particle center and the cell face (max. distance)
        # set the distance equal to the distance between particle center and the cell face
        elif distance_face_var >= distances_faces_max_i[face_index]:
            distance_face = distances_faces_max_i[face_index]
        # Else set distance to variable distance value
        else:
            distance_face = distance_face_var
        # Update list
        distances_faces.append(distance_face)
    # Calculate the difference between the aim distances and maximal distances
    delta_distances = [abs(a - b) for a, b in zip(distances_faces_max_i, distances_faces)]
    # Number of vertices
    n_verticies = len(cell_vertices_i)
    # Predefine list for new vertices
    cell_vertices_i_new = [(None, None, None) for _ in range(n_verticies)]
    # Consider each vertex of polyhedron
    for vertex_index in range(len(cell_vertices_i)):
        # Find every face that is connected with the considered vertex
        face_index = [i for i, subarray in enumerate(vertices_indices_i) if vertex_index in subarray]
        # Define move distances of vetex 
        distances_move = [delta_distances[i] for i in face_index]
        # Define move vectors of vetex
        vectors_move = np.array([cell_face_normals_i[i] for i in face_index])
        # Compute the displacement vector
        displacement = sum(distance * normal for distance, normal in zip(distances_move, vectors_move))
        # Calculate the new position coordinates of vertex
        verticies_coordinates = cell_vertices_i[vertex_index] - displacement
        # Update list
        cell_vertices_i_new[vertex_index] = verticies_coordinates
    # Convert the list of arrays to a 2D NumPy array
    cell_vertices_i_new = np.vstack(cell_vertices_i_new)
    
    return cell_vertices_i_new

def calc_volume_polyhedron(cell_vertices_i_new):
    """ Function to calculates the polyhedron volume based on the Tetrahedral Decomposition method"""
    # Polyhedron volume
    volume_polyhedron = 0
    # Perform Delaunay triangulation to get tetrahedra
    delaunay = Delaunay(cell_vertices_i_new)
    # Consider each tetrahedron
    for simplex in delaunay.simplices:
        # Get tetrahedron verticies
        verticies_tetrahedron = cell_vertices_i_new[simplex]
        # Calculate the tetrahedron volumes
        volume_tetrahedron = calc_tetrahedron_volume(verticies_tetrahedron)
        # Calculate the volume of the polyhedron by summing tetrahedron volumes
        volume_polyhedron += volume_tetrahedron

    return volume_polyhedron

def calc_tetrahedron_volume(tetra):
    """ Function to calculates the volume of a tetrahedron based on its vertices. """
    # Tetrahedron vertices
    a, b, c, d = tetra
    # Calculate tetrahedron volume
    volume_tetrahedron = abs(np.dot(a-d, np.cross(b-d, c-d))) / 6

    return volume_tetrahedron

def solver_grow_factor(fit_pars, holdup_end, volume_cube, diameters_0, volumes_cells):
    """ Solver to calculate the grow factor """
    # Variable grow factor
    grow_factor = fit_pars
    # Minimization value
    total_error = 0
    # New diameters
    diameters_new = diameters_0 * grow_factor
    # New volumes
    volumes_particles_new = (np.pi / 6) * diameters_new **(3)
    # Correct volumes
    for i in range(len(volumes_particles_new)):
        if volumes_particles_new[i] - volumes_cells[i] > 1e-9:
            volumes_particles_new[i] = volumes_cells[i]
    # New holdup                
    holdup_new = np.sum(volumes_particles_new) / volume_cube
    # 
    total_error = 100*abs(holdup_end - holdup_new) / holdup_end

    return total_error